#include "config.h"
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#include <stdlib.h>
#include <memory.h>
#include "assert.h"
#include "io.h"
#include "io_interface.h"
#include "io_internal.h"
#include "unix_defs.h"

int IOdumpVerbose = 0;

static int write_IOreformat_cue ARGS((IOFile iofile, int format_id));
static void server_register_format ARGS((IOFile iofile, IOFormat ioformat));
static void byte_swap ARGS((char *data, int size));
static IOFormat server_get_format ARGS((IOContext iocontext, void *buffer));
static void unstringify_field_type ARGS((char *type, char *buffer, int size));
/* 
 * this reference to establish_server_connection is a function pointer rather 
 * than an extern so that the server_interact.o file is not loaded unless
 * create_IOcontext() (which is also in that file) appears in the user program. * This avoid requiring the user to always link with -lnsl, -lsocket, etc 
 * when he might not otherwise need to.
 */
void (*establish_server_connection_ptr)ARGS((IOFile iofile, int do_fallback));

IOFormat
get_IOformat_by_name_IOcontext(iocontext, name)
IOContext iocontext;
char *name;
{
    IOFile iofile = (IOFile) iocontext;
    if (iofile->is_context) {
	char buffer[32];  /* server_ID size or larger */
	unstringify_field_type(name, buffer, sizeof(buffer));
	return get_format_IOcontext((IOContext) iofile, buffer);
    } else {
	return get_IOformat_by_name(iofile, name);
    }
}

IOFormat
get_IOformat_by_name(iofile, name)
IOFile iofile;
char *name;
{
    int i;
    if (iofile->status == OpenBeginRead) {
	/* if we're at the very begining of the file, read all formats to */
	/* preserve old behavior */
	(void) next_IOrecord_format(iofile);
	iofile->status = OpenForRead;
    }
    for (i = iofile->reg_format_count - 1; i >= 0; i--) {
	if ((iofile->format_list[i] != NULL) && /*entry null if reregistering*/
	    (strcmp(name, iofile->format_list[i]->format_name) == 0)) {
	    if (iofile->is_context && !iofile->format_list[i]->native_format) {
		/* in contexts, never match non-native formats by name */
		continue;
	    }
	    return iofile->format_list[i];
	}
    }
    return NULL;
}

static IOFormat
copy_ioformat_to_iofile(format, iofile)
IOFormat format;
IOFile iofile;
{
    IOFormat new_format = io_malloc(sizeof(IOFormatStruct));
    int i;

    memcpy(new_format, format, sizeof(IOFormatStruct));
    new_format->format_name = strdup(format->format_name);
    new_format->iofile = iofile;
    new_format->conversion = NULL;
    new_format->field_list = copy_field_list(format->field_list);
    new_format->var_list = NULL;
    generate_var_list(new_format);
    for (i=0 ; i< new_format->field_count; i++) {
	if (format->var_list[i].subformat != NULL) {
	    new_format->var_list[i].subformat = 
		copy_ioformat_to_iofile(format->var_list[i].subformat, iofile);
	}
    }
    if (iofile->reg_format_count == iofile->format_list_size) {
	expand_IOFile(iofile);
    }
    new_format->format_id = iofile->reg_format_count++;
    iofile->format_list[new_format->format_id] = new_format;
    
    return new_format;
}

IOFormat
get_format_IOcontext(iocontext, buffer)
IOContext iocontext;
void *buffer;
{
    IOFile iofile = (IOFile) iocontext;
    IOFormat new_format;
    int i;
    
    for (i = iofile->reg_format_count - 1; i >= 0; i--) {
	if (memcmp(buffer, iofile->format_list[i]->server_ID, 
		   sizeof(iofile->format_list[i]->server_ID)) == 0) {
	    return iofile->format_list[i];
	}
    }
    if (iofile->master_context == NULL) {
	new_format = server_get_format(iocontext, buffer);
    } else {
	/* we're a subcontext, get the format from the master */
	IOFormat master_format = 
	    get_format_IOcontext(((IOFile)iocontext)->master_context, buffer);

	/* copy it and enter it into our list */
	new_format = copy_ioformat_to_iofile(master_format, (IOFile)iocontext);
    }
    return new_format;
}

char **
get_subformat_names(field_list)
IOFieldList field_list;
{
    int name_count = 0;
    int field = 0;
    char **name_list = malloc(sizeof(char *));
    while (field_list[field].field_name != NULL) {
	char *base_type = base_data_type(field_list[field].field_type);
	if (str_to_data_type(base_type) == unknown_type) {
	    name_list = realloc(name_list, sizeof(char *) * (name_count +1));
	    name_list[name_count++] = base_type;
	} else {
	    free(base_type);
	}
	field++;
    }
    name_list = realloc(name_list, sizeof(char *) * (name_count +1));
    name_list[name_count] = NULL;
    if (name_count == 0) {
	free(name_list);
	name_list = NULL;
    }
    return name_list;
}

static
void
get_subformats_context(ioformat, format_list_p, format_count_p)
IOFormat ioformat;
IOFormat **format_list_p;
int *format_count_p;
{
    int field;
    for (field = 0; field < ioformat->field_count; field++) {
	IOFormat subformat = ioformat->var_list[field].subformat;
	if (subformat != NULL) {
	    get_subformats_context(subformat, format_list_p, format_count_p);
	    *format_list_p = realloc(*format_list_p, 
				     sizeof(IOFormat) * (*format_count_p +1));
	    (*format_list_p)[(*format_count_p)++] = subformat;
	}
    }
}
	    
IOFormat *
get_subformats_IOcontext(iocontext, buffer)
IOContext iocontext;
void *buffer;
{
    IOFormat ioformat = get_format_IOcontext(iocontext, buffer);
    if (ioformat == NULL) {
	int i, id_size = sizeof(((IOFile) iocontext)->format_list[0]->server_ID);
	char hex_buffer[64];
	for (i=0 ; i< id_size; i++) {
	    sprintf(&hex_buffer[2*i], "%02x", ((unsigned char*)buffer)[i]);
	}
	printf("format \"%s\" not registered with format server.\n",
	       hex_buffer);
	return NULL;
    }
    return get_subformats_IOformat(ioformat);
}

IOFormat *
get_subformats_IOformat(ioformat)
IOFormat ioformat;
{
    int format_count = 0;
    IOFormat *format_list = malloc(sizeof(IOFormat));
    get_subformats_context(ioformat, &format_list, &format_count);
    format_list = realloc(format_list, sizeof(IOFormat) * (format_count +2));
    format_list[format_count] = ioformat;
    format_list[format_count+1] = NULL;
    return format_list;
}

IOFormat
get_IOformat_by_index(iofile, index)
IOFile iofile;
int index;
{
    if (index < iofile->reg_format_count) {
	return iofile->format_list[index];
    } else {
	return NULL;
    }
}

static
IOFormat
new_IOFormat()
{
    IOFormat ioformat;
    ioformat = (IOFormat) io_malloc(sizeof(IOFormatStruct));
    ioformat->format_name = NULL;
    ioformat->conversion = NULL;
    ioformat->byte_reversal = 0;
    ioformat->pointer_size = 0;
    ioformat->IOversion = -1;
    ioformat->field_list = NULL;
    ioformat->var_list = NULL;
    ioformat->native_format = 1;
    memset(&ioformat->server_ID[0], 0, sizeof(ioformat->server_ID));
    return (ioformat);
}

extern
void
expand_IOFile(iofile)
IOFile iofile;
{
    int new_count = iofile->format_list_size + 10;
    int new_size = sizeof(IOFormat) * (new_count);
    int i;

    if (iofile->format_list != NULL) {
	iofile->format_list = (IOFormat *)
	    io_realloc((void *) iofile->format_list, (unsigned) new_size);
    } else {
	iofile->format_list = (IOFormat *) io_malloc((unsigned) new_size);
    }
    iofile->format_list_size = new_count;
    for(i=iofile->reg_format_count; i<iofile->format_list_size; i++) {
	iofile->format_list[i] = NULL;
    }
}

extern void
generate_var_list(ioformat)
IOFormat ioformat;
{
    IOVarInfoList new_var_list;
    IOFieldList field_list = ioformat->field_list;
    int field_count = ioformat->field_count;
    int field;

    if (ioformat->var_list) free(ioformat->var_list);
    new_var_list = (IOVarInfoList)
	io_malloc((size_t) sizeof(IOVarInfoStruct) * field_count);
    ioformat->var_list = new_var_list;
    for (field = 0; field < field_count; field++) {
	new_var_list[field].string = FALSE;
	new_var_list[field].var_array = FALSE;
	new_var_list[field].control_field = NULL;
	new_var_list[field].subformat = NULL;
	new_var_list[field].dimen1 = new_var_list[field].dimen2 = 0;
	new_var_list[field].data_type = 
	    array_str_to_data_type(field_list[field].field_type,
				   &new_var_list[field].dimen1, 
				   &new_var_list[field].dimen2);
	if (new_var_list[field].data_type == string_type) {
	    ioformat->variant = TRUE;
	    new_var_list[field].string = TRUE;
	} else {
	    int control_int = -1;
	    control_int =
		get_var_array_control(field_list[field].field_type,
				      field_list);
	    if (control_int != -1) {
		/* got a variant array */
		ioformat->variant = 1;
		new_var_list[field].var_array = TRUE;
		new_var_list[field].control_field =
		    get_IOfieldPtrFromList(field_list,
				     field_list[control_int].field_name);
	    }
	}
    }
}

IOFormat
register_IOcontext_format(format_name, field_list, iocontext)
char *format_name;
IOFieldList field_list;
IOContext iocontext;
{
    return register_IOrecord_format(format_name, field_list,
				    (IOFile) iocontext);
}

IOFormat
register_IOrecord_format(format_name, field_list, iofile)
char *format_name;
IOFieldList field_list;
IOFile iofile;
{
    return register_general_IOrecord_format(format_name, field_list, iofile,
					    sizeof(char*));
}

IOFormat
register_general_IOrecord_format(format_name, field_list, iofile, pointer_size)
char *format_name;
IOFieldList field_list;
IOFile iofile;
int pointer_size;
{
    IOFormat ioformat;
    IOFieldList new_field_list;
    int id, new_format_id = -1, field;
    int field_count = count_IOfield(field_list);

    if ((iofile->status != OpenNoHeader) && (iofile->status != OpenHeader)) {
	iofile->result = "Attempt to register IO format in inappropriate IOfile state";
	return NULL;
    }
    for (id = 0; id < iofile->reg_format_count; id++) {
	if (strcmp(format_name, iofile->format_list[id]->format_name) == 0) {
	    free_IOformat(iofile->format_list[id]);
	    new_format_id = id;
	}
    }

    ioformat = new_IOFormat();
    ioformat->format_name = strdup(format_name);
    ioformat->pointer_size = pointer_size;
    ioformat->IOversion = iofile->IOversion;
    if (new_format_id == -1) {
	if (iofile->reg_format_count == iofile->format_list_size) {
	    expand_IOFile(iofile);
	}
	ioformat->format_id = iofile->reg_format_count++;
    } else {
	ioformat->format_id = new_format_id;
	if ((iofile->status == OpenHeader) && (!iofile->is_context)) {
	    /* 
	     * already open and headers are away.  Cue the reader that
	     * we're reregistering a format.
	     */
	    if (!write_IOreformat_cue(iofile, new_format_id))
		return NULL;
	}
    }
    iofile->format_list[ioformat->format_id] = ioformat;
    ioformat->iofile = iofile;


    new_field_list = (IOFieldList) io_malloc((size_t) sizeof(IOField) *
					     (field_count + 1));
    ioformat->record_length = 0;
    ioformat->variant = FALSE;

    for (field = 0; field < field_count; field++) {
	int field_size = 0;
	if (get_var_array_control(field_list[field].field_type, field_list)
	    != -1) {
	    /* variant array, real_field_size is iofile->pointer_size */
	    field_size = ioformat->pointer_size;
	    if (field_size <= 0) {
		fprintf(stderr, "Field Size is not positive!  pointer size = %d\n",
			ioformat->pointer_size);
		exit(0);
	    }
	} else {
	    int dimen1 = 1, dimen2 = 1;
	    IOdata_type base_type;
	    base_type = array_str_to_data_type(field_list[field].field_type,
					       &dimen1, &dimen2);
	    if ((base_type != unknown_type) &&
		(field_list[field].field_size > 16)) {
		fprintf(stderr, "Field size for field %s in format %s is large, check to see if it is valid.\n",
		    field_list[field].field_name, ioformat->format_name);
	    }
	    field_size = field_list[field].field_size * dimen1 * dimen2;
	    if (field_size <= 0) {
		fprintf(stderr, "Field Size is not positive!  base= %d, dimen1=%d, dimen2=%d\n",
			field_list[field].field_size, dimen1, dimen2);
		exit(0);
	    }
	}
	
	ioformat->record_length = Max(ioformat->record_length,
				      field_list[field].field_offset +
				      field_size);
	new_field_list[field].field_name = strdup(field_list[field].field_name);
	new_field_list[field].field_type = strdup(field_list[field].field_type);
	new_field_list[field].field_size = field_list[field].field_size;
	new_field_list[field].field_offset = field_list[field].field_offset;
    }
    new_field_list[field_count].field_name = NULL;
    new_field_list[field_count].field_type = NULL;
    new_field_list[field_count].field_size = 0;
    new_field_list[field_count].field_offset = 0;

    if (ioformat->record_length == 0) {
	(void) fprintf(stderr, "Format %s has zero max length.\n",
		       format_name);
	(void) fprintf(stderr, "Format rejected.\n ");
	io_free(new_field_list);
	return NULL;
    }
    ioformat->field_count = field_count;
    ioformat->field_list = new_field_list;
    qsort(ioformat->field_list, field_count, sizeof(IOField),
	  field_offset_compar);
    generate_var_list(ioformat);

    for (field = 0; field < field_count; field++) {
	if (ioformat->var_list[field].string == TRUE) {
	    ioformat->variant = TRUE;
	} else {
	    char *base_type =
		base_data_type(ioformat->field_list[field].field_type);
	    IOFormat subformat;
	    if (str_to_data_type(base_type) == unknown_type) {
		 subformat = get_IOformat_by_name(iofile, base_type);
		 if (ioformat->var_list[field].var_array == TRUE) {
		     /* got a variant array */
		     ioformat->variant = 1;
		 } else {
		     if (subformat != NULL) {
			 ioformat->variant |= subformat->variant;
		     }
		 }
		 if (subformat == NULL) {
		     (void) fprintf(stderr, "Field \"%s\" base type \"%s\" is not a simple type or registered format name.\n",
				    ioformat->field_list[field].field_name, 
				    base_type);
		     (void) fprintf(stderr, "Format rejected.\n ");
		     free_IOformat(ioformat);
		     return NULL;
		 } else {
		     ioformat->var_list[field].subformat = subformat;
		 }
	    }
	    free(base_type);
	}
    }
    for (field = 1; field < field_count; field++) {
	int last_field_end = new_field_list[field - 1].field_offset +
	new_field_list[field - 1].field_size;
	if (ioformat->var_list[field - 1].var_array) {
	    last_field_end = new_field_list[field - 1].field_offset +
		ioformat->pointer_size;
	}
	if (new_field_list[field].field_offset < last_field_end) {
	    (void) fprintf(stderr, "Fields \"%s\" and \"%s\" in format \"%s\" overlap!  Format rejected.\n",
			   new_field_list[field - 1].field_name,
			   new_field_list[field].field_name,
			   format_name);
	    io_free(new_field_list);
	    return NULL;
	}
    }
    if (iofile->is_context) {
	/* register format with server */
	server_register_format(iofile, ioformat);
    } else if (iofile->status == OpenHeader) {
	/* already open and headers are away.  Write the format now */
	if (!write_IOformat(iofile, ioformat))
	    return NULL;
    }
    return ioformat;
}

extern
int
count_IOfield(list)
IOFieldList list;
{
    int i = 0;
    while (list[i].field_name != NULL) {
	i++;
    }
    return i;
}

extern
int
struct_size_IOfield(iofile, list)
IOFile iofile;
IOFieldList list;
{
    int i = 0;
    int struct_size = 0;
    while (list[i].field_name != NULL) {
	int field_size = 0;
	if (get_var_array_control(list[i].field_type, list) != -1) {
	    /* variant array, real_field_size is ioformat->pointer_size */
	    if ((iofile == NULL) || (iofile->pointer_size == 0)){
		field_size = sizeof(char *);
	    } else {
		field_size = iofile->pointer_size;
	    }
	} else {
	    if (iofile && (iofile->IOversion > 3)) {
		int dimen1 = 1, dimen2 = 1;
		array_str_to_data_type(list[i].field_type, &dimen1, &dimen2);
		field_size = list[i].field_size * dimen1 * dimen2;
	    } else {
		field_size = list[i].field_size;
	    }
	}
	assert(field_size > 0);
	struct_size = Max(struct_size,
			  (list[i].field_offset + field_size));
	i++;
    }
    return struct_size;
}

extern
int
struct_size_field_list(list, pointer_size)
IOFieldList list;
int pointer_size;
{
    int i = 0;
    int struct_size = 0;
    while (list[i].field_name != NULL) {
	int field_size = 0;
	if (get_var_array_control(list[i].field_type, list) != -1) {
	    /* variant array, real_field_size is ioformat->pointer_size */
	    field_size = pointer_size;
	} else {
	    int dimen1 = 1, dimen2 = 1;
	    array_str_to_data_type(list[i].field_type, &dimen1, &dimen2);
	    field_size = list[i].field_size * dimen1 * dimen2;
	}
	assert(field_size > 0);
	struct_size = Max(struct_size,
			  (list[i].field_offset + list[i].field_size));
	i++;
    }
    return struct_size;
}

extern
IOFieldList
field_list_of_IOformat(format)
IOFormat format;
{
    return format->field_list;
}

extern
int
compare_field_lists(list1, list2)
IOFieldList list1, list2;
{
    int i = 0;
    do {
	if ((strcmp(list1[i].field_name, list2[i].field_name) != 0) ||
	    (list1[i].field_size != list2[i].field_size) ||
	    (list1[i].field_offset != list2[i].field_offset) ||
	    !IO_field_type_eq(list1[i].field_type, list2[i].field_type)) {
	    return 1;
	}
	i++;
    } while ((list1[i].field_name != NULL) || (list2[i].field_name != NULL));

    return 0;
}

/* 
 * Max field list returns the "maximum" of two field lists. 
 * Basically, it prefers list1.  If all fields in list2 are in list 1, and 
 * if no fields in list2 are bigger than the corresponding field in
 * list1, the maximum returned is list1.  If there is a field in list2 
 * that is bigger than the corresponding field in list1, we'll still
 * use the ordering and even try to use the offsets in list1.  If
 * there's actually a field in list2 that isn't in list1, then most
 * bets are off as to exactly what will happen.  Basically the
 * resulting list will contain all fields, but we'll use the offsets
 * from their original lists to establish an ordering and then bump
 * the offsets up as necessary to ensure no overlap.  The result is 
 * likely a jumble that isn't necessarily terribly close to what was in 
 * either list...  Bummer for performance perhaps, but being smarter is
 * hard. 
 */
extern
IOFieldList
max_field_lists(list1, list2)
IOFieldList list1, list2;
{
    IOFieldList max_field_list = NULL;
    IOFieldList tlist2;
    int field_count1 = count_IOfield(list1);
    int field_count2 = count_IOfield(list2);
    int max_field_count = 0;
    int i, j;

    /* make a copy of list2 that we can scratch on */
    tlist2 = (IOFieldList) io_malloc((size_t) sizeof(IOField) * field_count2);
    memcpy(tlist2, list2, sizeof(IOField) * field_count2);

    max_field_list = (IOFieldList) io_malloc((size_t) sizeof(IOField) *
				      (field_count1 + field_count2 + 1));

    for (i = 0; i < field_count1; i++) {
	/* for each field in list1 */
	for (j = 0; j < field_count2; j++) {
	    if (tlist2[j].field_name &&
	      (strcmp(list1[i].field_name, tlist2[j].field_name) == 0)) {
		break;
	    }
	}
	if (j == field_count2) {
	    /* field exists only in 1, use it as is */
	    max_field_list[max_field_count].field_name =
		strdup(list1[i].field_name);
	    max_field_list[max_field_count].field_type =
		strdup(list1[i].field_type);
	    max_field_list[max_field_count].field_size =
		list1[i].field_size;
	    max_field_list[max_field_count].field_offset =
		list1[i].field_offset;
	    max_field_count++;
	} else {
	    /* field list in both.  Check types and take max size */
	    if (!IO_field_type_eq(list1[i].field_type, tlist2[j].field_type)) {
		/* Yikes!  Types are bad!  Serious problem */
		io_free(tlist2);
		io_free(max_field_list);
		return NULL;
	    }
	    max_field_list[max_field_count].field_name =
		strdup(list1[i].field_name);
	    max_field_list[max_field_count].field_type =
		strdup(list1[i].field_type);
	    max_field_list[max_field_count].field_size =
		Max(list1[i].field_size, tlist2[j].field_size);
	    /* tentative offset */
	    max_field_list[max_field_count].field_offset =
		list1[i].field_offset;
	    max_field_count++;
	    /* mark field in list 2 as used */
	    tlist2[j].field_name = NULL;
	}
    }
    for (j = 0; j < field_count2; j++) {
	if (tlist2[j].field_name != NULL) {
	    /* didn't mark this field as used.  Must only be in list2 */
	    max_field_list[max_field_count].field_name =
		strdup(tlist2[j].field_name);
	    max_field_list[max_field_count].field_type =
		strdup(tlist2[j].field_type);
	    max_field_list[max_field_count].field_size =
		tlist2[j].field_size;
	    max_field_list[max_field_count].field_offset =
		tlist2[j].field_offset;
	    max_field_count++;
	}
    }
    io_free(tlist2);

    max_field_list[max_field_count].field_name = NULL;
    max_field_list[max_field_count].field_type = NULL;
    max_field_list[max_field_count].field_size = 0;
    max_field_list[max_field_count].field_offset = 0;

    /* sort the max list by offset now */
    qsort(max_field_list, max_field_count, sizeof(IOField),
	  field_offset_compar);
    for (i = 1; i < max_field_count; i++) {
	int last_field_end = max_field_list[i - 1].field_offset +
	max_field_list[i - 1].field_size;
	if (max_field_list[i].field_offset < last_field_end) {
	    /* damn, got some overlap.  Push this field up to fix. */
	    max_field_list[i].field_offset = last_field_end;
	}
    }
    return max_field_list;
}

extern int
min_align_size(size)
int size;
{
    int align_size = 8;		/* conservative on current machines */
    switch (size) {
    case 7:
    case 6:
    case 5:
    case 4:
	align_size = 4;
	break;
    case 3:
	align_size = 2;
	break;
    case 2:
    case 1:
	align_size = size;
	break;
    }
    return align_size;
}

extern void
force_align_field_list(field_list, pointer_size)
IOFieldList field_list;
int pointer_size;
{
    int i;
    for (i = 0; field_list[i].field_name != NULL; i++) {
	int align_size = min_align_size(field_list[i].field_size);
	if (i > 0) {
	    int last_field_end = field_list[i - 1].field_offset +
	    field_list[i - 1].field_size;
	    if (field_list[i].field_offset < last_field_end) {
		/* damn, got some overlap.  Push this field up to fix. */
		field_list[i].field_offset = last_field_end;
	    }
	}
	if ((field_list[i].field_offset % align_size) != 0) {
	    /* field is mis-aligned, move it up...  */
	    int offset = field_list[i].field_offset + align_size;
	    /* adjust down to the boundary */
	    offset -= (field_list[i].field_offset % align_size);
	    field_list[i].field_offset = offset;
	}
    }
}

extern
IOFieldList
copy_field_list(list)
IOFieldList list;
{
    int field_count = count_IOfield(list);
    IOFieldList new_field_list;
    int field;

    new_field_list = (IOFieldList) io_malloc((size_t) sizeof(IOField) *
					     (field_count + 1));
    for (field = 0; field < field_count; field++) {
	new_field_list[field].field_name = strdup(list[field].field_name);
	new_field_list[field].field_type = strdup(list[field].field_type);
	new_field_list[field].field_size = list[field].field_size;
	new_field_list[field].field_offset = list[field].field_offset;
    }
    new_field_list[field_count].field_name = NULL;
    new_field_list[field_count].field_type = NULL;
    new_field_list[field_count].field_offset = 0;
    new_field_list[field_count].field_size = 0;
    return new_field_list;
}

extern
void
free_field_list(list)
IOFieldList list;
{
    int i = 0;
    while (list[i].field_name != NULL) {
	io_free(list[i].field_name);
	io_free(list[i].field_type);
	i++;
    }
    io_free(list);
}

IOFormat
reregister_IOformat(other_format, iofile)
IOFormat other_format;
IOFile iofile;
{
    IOFormat ioformat;
    IOFieldList new_field_list, field_list = other_format->field_list;
    int id, field, new_format_id = -1;
    char *format_name = other_format->format_name;

    if ((iofile->status != OpenNoHeader) && (iofile->status != OpenHeader)) {
	iofile->result = "Attempt to register IO format in inappropriate IOfile state";
	return NULL;
    }
    for (id = 0; id < iofile->reg_format_count; id++) {
	if (strcmp(format_name, iofile->format_list[id]->format_name) == 0) {
	    io_free(iofile->format_list[id]->field_list);
	    io_free(iofile->format_list[id]);
	    new_format_id = id;
	}
    }

    ioformat = new_IOFormat();
    ioformat->format_name = format_name;
    ioformat->pointer_size = sizeof(char*);
    ioformat->IOversion = iofile->IOversion;
    if (new_format_id == -1) {
	if (iofile->reg_format_count == iofile->format_list_size) {
	    expand_IOFile(iofile);
	}
	ioformat->format_id = iofile->reg_format_count++;
    } else {
	ioformat->format_id = new_format_id;
	if (iofile->status == OpenHeader) {
	    /* 
	     * already open and headers are away.  Cue the reader that
	     * we're reregistering a format. 
	     */
	    if (!write_IOreformat_cue(iofile, new_format_id))
		return NULL;
	}
    }
    iofile->format_list[ioformat->format_id] = ioformat;
    ioformat->iofile = iofile;


    new_field_list = (IOFieldList) io_malloc((size_t) sizeof(IOField) *
					(other_format->field_count + 1));
    ioformat->record_length = other_format->record_length;
    ioformat->variant = other_format->variant;
    for (field = 0; field < other_format->field_count; field++) {
	new_field_list[field].field_name = strdup(field_list[field].field_name);
	new_field_list[field].field_type = strdup(field_list[field].field_type);
	new_field_list[field].field_size = field_list[field].field_size;
	new_field_list[field].field_offset = field_list[field].field_offset;
    }
    new_field_list[other_format->field_count].field_name = NULL;
    new_field_list[other_format->field_count].field_type = NULL;
    new_field_list[other_format->field_count].field_size = 0;
    new_field_list[other_format->field_count].field_offset = 0;

    ioformat->field_count = other_format->field_count;
    ioformat->field_list = new_field_list;
    qsort(ioformat->field_list, ioformat->field_count, sizeof(IOField),
	  field_offset_compar);
    if (iofile->status == OpenHeader) {
	/* already open and headers are away.  Write the format now */
	if (!write_IOformat(iofile, ioformat))
	    return NULL;
    }
    return ioformat;
}

extern
IOFormat
duplicate_IOrecord_format(src, iofile)
IOFormat src;
IOFile iofile;
{
    IOFormat ioformat;
    IOFieldList new_field_list;
    int id, field;

    if ((iofile->status != OpenNoHeader) && (iofile->status != OpenHeader)) {
	iofile->result = "Attempt to register IO format in inappropriate IOfile state";
	return NULL;
    }
    for (id = 0; id < iofile->reg_format_count; id++) {
	if (strcmp(src->format_name,
		   iofile->format_list[id]->format_name) == 0) {
	    iofile->result = "Attempt to register duplicate format name";
	    return NULL;
	}
    }
    ioformat = new_IOFormat();
    ioformat->format_name = strdup(src->format_name);
    ioformat->pointer_size = src->pointer_size;
    ioformat->IOversion = src->IOversion;
    ioformat->native_format = src->native_format;
    if (iofile->reg_format_count == iofile->format_list_size) {
	expand_IOFile(iofile);
    }
    ioformat->format_id = iofile->reg_format_count++;
    iofile->format_list[ioformat->format_id] = ioformat;
    ioformat->iofile = iofile;

    new_field_list = (IOFieldList)
	io_malloc((size_t) sizeof(IOField) * (src->field_count + 1));
    ioformat->record_length = 0;
    ioformat->variant = src->variant;;
    for (field = 0; field < src->field_count; field++) {
	ioformat->record_length = Max(ioformat->record_length,
				    src->field_list[field].field_offset +
				      src->field_list[field].field_size);
	new_field_list[field].field_name =
	    strdup(src->field_list[field].field_name);
	new_field_list[field].field_type =
	    strdup(src->field_list[field].field_type);
	new_field_list[field].field_size = src->field_list[field].field_size;
	new_field_list[field].field_offset =
	    src->field_list[field].field_offset;
    }
    new_field_list[src->field_count].field_name = NULL;
    new_field_list[src->field_count].field_type = NULL;
    new_field_list[src->field_count].field_size = 0;
    new_field_list[src->field_count].field_offset = 0;
    ioformat->field_count = src->field_count;
    ioformat->field_list = new_field_list;
    return ioformat;
}

extern
int
duplicate_IOfile_format(srcfile, iofile)
IOFile srcfile, iofile;
{
    int i;
    IOFormat ioformat;
    for (i = 0; i < srcfile->reg_format_count; i++) {
	ioformat = duplicate_IOrecord_format(srcfile->format_list[i],
					     iofile);
	if (ioformat == NULL)
	    return -1;
    }
    iofile->byte_reversal = srcfile->byte_reversal;
    iofile->architecture = strdup(srcfile->architecture);
    return 0;
}

extern
IOFormat
next_IOrecord_format(iofile)
IOFile iofile;
{
    if ((iofile->status != OpenForRead) && (iofile->status != OpenBeginRead))
	return NULL;

    if (iofile->read_ahead == FALSE) {
	(void) next_IOrecord_type(iofile);
    }
    while (iofile->next_record_type != IOdata) {
	switch (iofile->next_record_type) {
	case IOcomment:
	    (void) read_comment_IOfile(iofile);
	    (void) next_IOrecord_type(iofile);
	    break;
	case IOformat:
	    (void) read_format_IOfile(iofile);
	    (void) next_IOrecord_type(iofile);
	    break;
	default:
	    return NULL;
	}
    }
    return iofile->next_record_format;
}

extern
int
index_of_IOformat(format)
IOFormat format;
{
    return format->format_id;
}

extern
char *
name_of_IOformat(format)
IOFormat format;
{
    return format->format_name;
}

extern
int
pointer_size_of_IOformat(format)
IOFormat format;
{
    return format->pointer_size;
}

extern
IOFile
iofile_of_IOformat(format)
IOFormat format;
{
    return format->iofile;
}


static int
write_IOreformat_cue(iofile, format_id)
IOFile iofile;
int format_id;
{
    FILE_INT file_format_id = format_id;
    FILE_INT code = REREGISTERED_FORMAT_FOLLOWS;

    if (!put_AtomicInt(iofile, &code))
	return 0;
    if (!put_AtomicInt(iofile, &file_format_id))
	return 0;
    return 1;
}


#define WRITEV_FIELD_COUNT_LIMIT 10
#define WRITE_BLOCK_SIZE (6 + WRITEV_FIELD_COUNT_LIMIT * 4)
#define  WRITE_VEC_SIZE (2 + 3 * WRITEV_FIELD_COUNT_LIMIT)
extern int
write_IOformat(iofile, ioformat)
IOFile iofile;
IOFormat ioformat;
{
    FILE_INT write_block[WRITE_BLOCK_SIZE];
    internal_iovec write_vec[WRITE_VEC_SIZE];
    FILE_INT name_length, field_count, record_length, format_id, pointer_size;
    int field;
    FILE_INT code = FORMAT_FOLLOWS;
    int rec_len_index = 4;

    int write_vec_index = 0;
    int write_block_index = 0;

    name_length = strlen(ioformat->format_name);
    format_id = ioformat->format_id;
    field_count = ioformat->field_count;
    record_length = ioformat->record_length;
    pointer_size = ioformat->pointer_size;
    write_block[0] = code;
    write_block[1] = name_length;
    write_block[2] = format_id;
    write_block[3] = field_count;
    if (iofile->IOversion >= 5) {
	write_block[4] = pointer_size;
	rec_len_index = 5;
    }
    write_block[rec_len_index] = record_length;

    write_block_index = rec_len_index + 1;

    write_vec[0].iov_base = &write_block[0];
    write_vec[0].iov_len  = write_block_index * sizeof(FILE_INT);
    write_vec[0].iov_offset = 0;
    write_vec[1].iov_base = ioformat->format_name;
    write_vec[1].iov_len = name_length + 1;
    write_vec[1].iov_offset = 0;

    write_vec_index = 2;
    field = 0;
    while(field < ioformat->field_count) {
	while ((write_vec_index + 3 < WRITE_VEC_SIZE) && 
	       (field < ioformat->field_count)){
	    IOField *iofield = &(ioformat->field_list[field]);
	    /* field_name_len */
	    int field_name_len = strlen(iofield->field_name);
	    /* field_type_len */
	    int field_type_len = strlen(iofield->field_type);
	    write_block[write_block_index] = field_name_len;
	    write_block[write_block_index+1] = field_type_len;
	    /* field_size */
	    write_block[write_block_index+2] = iofield->field_size;
	    /* field_offset */
	    write_block[write_block_index+3] = iofield->field_offset;

	    write_vec[write_vec_index].iov_base =
		&write_block[write_block_index];
	    write_block_index += 4;

	    write_vec[write_vec_index].iov_len  = 4*sizeof(FILE_INT);
	    write_vec[write_vec_index].iov_offset = 0;
	    write_vec[write_vec_index+1].iov_base = iofield->field_name;
	    write_vec[write_vec_index+1].iov_len = field_name_len + 1;
	    write_vec[write_vec_index+1].iov_offset = 0;
	    write_vec[write_vec_index+2].iov_base = iofield->field_type;
	    write_vec[write_vec_index+2].iov_len = field_type_len + 1;
	    write_vec[write_vec_index+2].iov_offset = 0;
	    write_vec_index += 3;
	    field++;
	}
	if (AtomicWriteV(iofile, &write_vec[0], write_vec_index) !=
	    write_vec_index) {
	    return 0;
	}
	write_vec_index = 0;
	write_block_index = 0;
    }
    return 1;
}

IOFormat
read_format_IOfile(iofile)
IOFile iofile;
{
    FILE_INT name_length;
    FILE_INT field_count;
    FILE_INT record_length;
    FILE_INT format_id;
    IOFormat ioformat = new_IOFormat();
    IOVarInfoList new_var_list;
    int field;

    iofile->status = OpenForRead;
    ioformat->byte_reversal = iofile->byte_reversal;
    ioformat->IOversion = iofile->IOversion;
    ioformat->pointer_size = iofile->pointer_size;
    ioformat->native_format = 0;
    if ((iofile->IOversion > 1) && (next_IOrecord_type(iofile) != IOformat)) {
	iofile->result = "Next record type is not a format";
	return NULL;
    }
    reset_read_ahead(iofile);
    if (!get_AtomicInt(iofile, &name_length)) return NULL;
    if (iofile->IOversion >= 3) {
	if (!get_AtomicInt(iofile, &format_id)) return NULL;
    } else {
	format_id = iofile->reg_format_count;
    }
    if (!get_AtomicInt(iofile, &field_count)) return NULL;
    if (iofile->IOversion >= 5) {
	FILE_INT pointer_size;
	if (!get_AtomicInt(iofile, &pointer_size)) return NULL;
	ioformat->pointer_size = pointer_size;
    }
    if (!get_AtomicInt(iofile, &record_length)) return NULL;

    ioformat->format_name = io_malloc(name_length + 1);
    ioformat->field_count = field_count;
    ioformat->variant = FALSE;
    ioformat->record_length = record_length;
    ioformat->conversion = NULL;
    AtomicRead(iofile->file_id, ioformat->format_name, name_length + 1,iofile);
    ioformat->field_list = (IOFieldList) io_malloc(sizeof(IOField) *
					    (ioformat->field_count + 1));
    new_var_list = (IOVarInfoList)
	io_malloc((size_t) sizeof(IOVarInfoStruct) * field_count);
    ioformat->var_list = new_var_list;

    for (field = 0; field < ioformat->field_count; field++) {
	FILE_INT field_name_len;
	FILE_INT field_type_len;
	FILE_INT field_size;
	FILE_INT field_offset;
	IOField *iofield = &(ioformat->field_list[field]);

	if (!get_AtomicInt(iofile, &field_name_len)) return NULL;
	if (!get_AtomicInt(iofile, &field_type_len)) return NULL;
	if (!get_AtomicInt(iofile, &field_size)) return NULL;
	if (!get_AtomicInt(iofile, &field_offset)) return NULL;

	iofield->field_size = field_size;
	iofield->field_offset = field_offset;
	iofield->field_name = io_malloc(field_name_len + 1);
	iofield->field_type = io_malloc(field_type_len + 1);
	if (AtomicRead(iofile->file_id, iofield->field_name, 
		       field_name_len + 1, iofile) != field_name_len + 1) {
	    return NULL;
	}
	if (AtomicRead(iofile->file_id, iofield->field_type, 
		       field_type_len + 1, iofile) != field_type_len + 1) {
	    return NULL;
	}
    }
    /* 
     * now that we've read them all, fill in variant fields
     */
    for (field = 0; field < ioformat->field_count; field++) {
	IOField *iofield = &(ioformat->field_list[field]);

	new_var_list[field].string = FALSE;
	new_var_list[field].var_array = FALSE;
	new_var_list[field].control_field = NULL;
	new_var_list[field].subformat = NULL;
	new_var_list[field].data_type = 
	    array_str_to_data_type(iofield->field_type, 
				   &new_var_list[field].dimen1, 
				   &new_var_list[field].dimen2);
	if (new_var_list[field].data_type == string_type) {
	    ioformat->variant = TRUE;
	    new_var_list[field].string = TRUE;
	} else {
	    int control_int = -1;
	    char *base_type = base_data_type(iofield->field_type);
	    IOFormat subformat = get_IOformat_by_name(iofile, base_type);
	    new_var_list[field].string = FALSE;
	    control_int =
		get_var_array_control(ioformat->field_list[field].field_type,
				      ioformat->field_list);
	    free(base_type);
	    if (control_int != -1) {
		/* got a variant array */
		ioformat->variant = 1;
		new_var_list[field].var_array = TRUE;
		new_var_list[field].control_field =
		    get_IOfieldPtrFromList(ioformat->field_list,
			   ioformat->field_list[control_int].field_name);
		new_var_list[field].control_field->byte_swap =
		    ioformat->byte_reversal;
	    } else {
		if (subformat != NULL) {
		    ioformat->variant |= subformat->variant;
		}
	    }
	    ioformat->var_list[field].subformat = subformat;
	}
    }
    ioformat->field_list[ioformat->field_count].field_name = NULL;
    ioformat->field_list[ioformat->field_count].field_type = NULL;
    ioformat->field_list[ioformat->field_count].field_offset = 0;
    ioformat->field_list[ioformat->field_count].field_size = 0;

    if (format_id >= iofile->format_list_size) {
	expand_IOFile(iofile);
    }
    if (format_id > iofile->reg_format_count) {
	iofile->result = "Internal error. skipped format ids.";
	return NULL;
    }
    if (format_id == iofile->reg_format_count) {
	/* new format came in */
	iofile->reg_format_count++;
    }
    ioformat->format_id = format_id;

    if (iofile->format_list[ioformat->format_id] != NULL) {
	free_IOformat(iofile->format_list[ioformat->format_id]);
    }
    iofile->format_list[ioformat->format_id] = ioformat;
    ioformat->iofile = iofile;
    return ioformat;
}

extern void
dump_IOFormat(ioformat)
IOFormat ioformat;
{
    int index;
    printf("\tFormat %d \"%s\"; size = %d; Field_Count = %d; Endian = %d; Pointer size = %d; File Version = %d\n",
     ioformat->format_id, ioformat->format_name, ioformat->record_length,
	   ioformat->field_count, ioformat->byte_reversal,
	   ioformat->pointer_size, ioformat->IOversion);
    for (index = 0; index < ioformat->field_count; index++) {
	printf("\t\t%s \"%s\"; size = %d; offset = %d\n",
	       ioformat->field_list[index].field_name,
	       ioformat->field_list[index].field_type,
	       ioformat->field_list[index].field_size,
	       ioformat->field_list[index].field_offset);
    }
    dump_IOConversion(ioformat->conversion);

}

static void
dump_IOfield ARGS((IOFile iofile, IOFormat ioformat, int index, void *data,
		   void *string_base));

extern void
dump_raw_IOrecord(iofile, ioformat, data)
IOFile iofile;
IOFormat ioformat;
void *data;
{
    int index;
    if (IOdumpVerbose)
	printf("Record type %s :", ioformat->format_name);
    for (index = 0; index < ioformat->field_count; index++) {
	dump_IOfield(iofile, ioformat, index, data, data);
    }
    printf("\n");
}


extern int
sdump_value(str, field_type, field_size, field_offset, iofile, data,
	    string_base, byte_reversal)
char *str;
char *field_type;
int field_size;
int field_offset;
IOFile iofile;
void *data;
void *string_base;
int byte_reversal;
{
    IOgetFieldStruct descr;
    descr.offset = field_offset;
    descr.size = field_size;
    descr.data_type = str_to_data_type(field_type);
    descr.byte_swap = byte_reversal;

    if (descr.data_type == integer_type) {
	if (field_size <= sizeof(long)) {
	    long tmp = get_IOlong(&descr, data);
	    sprintf(str, "%ld", tmp);
	} else if (field_size == 2 * sizeof(long) && field_size == 8) {
	    unsigned long low_long;
	    long high_long;
	    get_IOlong8(&descr, data, &low_long, &high_long);
	    if (high_long == 0) {
		sprintf(str, "%ld", low_long);
	    } else {
		sprintf(str, "0x%lx%08lx", high_long, low_long);
	    }
	} else if (field_size > sizeof(long)) {
	    sprintf(str, "+long int+");
	} else {
	    sprintf(str, "+int size %d+", field_size);
	}
    } else if (descr.data_type == unsigned_type) {
	if (field_size <= sizeof(unsigned long)) {
	    unsigned long tmp = get_IOulong(&descr, data);
	    sprintf(str, "%lu", tmp);
	} else if (field_size == 2 * sizeof(long) && field_size == 8) {
	    unsigned long low_long, high_long;
	    get_IOulong8(&descr, data, &low_long, &high_long);
	    if (high_long == 0) {
		sprintf(str, "%lu", low_long);
	    } else {
		sprintf(str, "0x%lx%08lx", high_long, low_long);
	    }
	} else if (field_size > sizeof(long)) {
	    sprintf(str, "+ulong int+");
	} else {
	    sprintf(str, "+uint size %u+", field_size);
	}
    } else if (descr.data_type == enumeration_type) {
	sprintf(str, "%u", *(int *) ((char *) data + field_offset));
    } else if (descr.data_type == float_type) {
	if (field_size == sizeof(float)) {
	    float tmp = get_IOfloat(&descr, data);
	    sprintf(str, "%g", tmp);
	} else if (field_size == sizeof(double)) {
	    double tmp = get_IOdouble(&descr, data);
	    sprintf(str, "%g", tmp);
#if SIZEOF_LONG_DOUBLE != 0 && SIZEOF_LONG_DOUBLE != SIZEOF_DOUBLE
	} else if (field_size == sizeof(long double)) {
	    long double tmp;
	    memcpy(&tmp, (float *) ((char *) data + field_offset),
		   sizeof(double));
	    sprintf(str, "%Lg", tmp);
#endif
	} else {
	    if (field_size < sizeof(float)) {
		sprintf(str, "+tiny float+");
	    } else if (field_size > sizeof(double)) {
		sprintf(str, "+big float+");
	    } else {
		sprintf(str, "+float size %u+", field_size);
	    }
	}
    } else if (descr.data_type == char_type) {
	sprintf(str, "%c", *(char *) ((char *) data + field_offset));
    } else if (descr.data_type == string_type) {
	char *tmp_str = get_IOstring_base(&descr, data, string_base);
	if (tmp_str == 0) {
	    sprintf(str, "\"\"");
	} else {
	    sprintf(str, "\"%s\"", tmp_str);
	}
    } else if (strcmp(field_type, "R3vector") == 0) {
	sprintf(str, "(%g, %g, %g)", *(double *) ((char *) data + field_offset),
	     *(double *) ((char *) data + field_offset + sizeof(double)),
	*(double *) ((char *) data + field_offset + 2 * sizeof(double)));
    } else {
	return 0;
    }
    return 1;
}

static void
dump_value(field_type, field_size, field_offset, iofile, data, string_base, byte_reversal)
char *field_type;
int field_size;
int field_offset;
IOFile iofile;
void *data;
void *string_base;
int byte_reversal;
{
    char buf[1024];		/* surely big enough for a single value? */
    IOFormat ioformat;
    if (sdump_value(buf, field_type, field_size, field_offset, iofile, data,
		    string_base, byte_reversal)) {
	printf("%s", buf);
    } else {
	char *typ = base_data_type(field_type);
	if ((ioformat = get_IOformat_by_name(iofile, typ)) != NULL) {
	    int index;
	    printf("{ ");
	    for (index = 0; index < ioformat->field_count; index++) {
		dump_IOfield(iofile, ioformat, index,
		   (void *) ((char *) data + field_offset), string_base);
	    }
	    printf("} ");
	} else {
	    printf("Unknown type");
	}
	free(typ);
    }
}

static void
dump_IOfield(iofile, ioformat, field, data, string_base)
IOFile iofile;
IOFormat ioformat;
int field;
void *data;
void *string_base;
{
    IOFieldList iofield = &ioformat->field_list[field];
    IOVarInfoList iovar = &ioformat->var_list[field];
    int field_offset = iofield->field_offset;
    int field_size = iofield->field_size;
    char *field_type = iofield->field_type;
    char *field_name = iofield->field_name;

    char *left_paren = NULL;
    if (IOdumpVerbose)
	printf("%s = ", field_name);
    if ((left_paren = strchr(field_type, '[')) == NULL) {
	dump_value(field_type, field_size, field_offset, iofile, data,
		   string_base,ioformat->byte_reversal);
    } else if (strchr(left_paren + 1, '[') == NULL) {
	/* single dimen array */
	int dimension = 0;
	char sub_type[64];
	int sub_field_size, offset = iofield->field_offset;
	printf("{ ");
	*left_paren = 0;
	strcpy(sub_type, field_type);
	*left_paren = '[';
	if (sscanf(left_paren + 1, "%d]", &dimension) != 1) {
	    if (iovar->var_array != TRUE) {
		fprintf(stderr, "Couldn't parse array size in \"%s\"\n",
			field_type);
		return;
	    } else {
		IOgetFieldStruct descr;
		descr.offset = iofield->field_offset;
		descr.size = ioformat->pointer_size;
		descr.data_type = integer_type;
		descr.byte_swap = ioformat->byte_reversal;

		dimension = get_IOlong(iovar->control_field, data);
		offset = get_IOlong(&descr, data);
		data = (char*)string_base + offset;
		offset = 0;
	    }
	    sub_field_size = iofield->field_size;
	} else {
	    /* normal internal array */
	    if (ioformat->IOversion <= 3) {
		sub_field_size = iofield->field_size / dimension;
	    } else {
		sub_field_size = iofield->field_size;
	    }
	}
	for (; dimension > 0; dimension--) {
	    dump_value(sub_type, sub_field_size, offset, iofile, data,
		       string_base, ioformat->byte_reversal);
	    offset += sub_field_size;
	    if (dimension != 1)
		printf(", ");
	}
	printf("}");
    } else {
	/* double dimen array */
	int dimension1 = 0;
	int dimension2 = 0;
	char sub_type[64];
	int sub_field_size, offset = iofield->field_offset;
	printf("{ ");
	*left_paren = 0;
	strcpy(sub_type, field_type);
	if (sscanf(left_paren + 1, "%d][%d]", &dimension1, &dimension2) != 2) {
	    *left_paren = '[';
	    fprintf(stderr, "Couldn't parse array size in \"%s\"\n",
		    field_type);
	    return;
	}
	*left_paren = '[';
	if (ioformat->IOversion <= 3) {
	    sub_field_size = iofield->field_size / (dimension1 * dimension2);
	} else {
	    sub_field_size = iofield->field_size;
	}
	for (; dimension2 > 0; dimension2--) {
	    int i = 0;
	    printf("\n\t{ ");
	    for (; i < dimension1; i++) {
		dump_value(sub_type, sub_field_size, offset, iofile, data,
			   string_base, ioformat->byte_reversal);
		offset += sub_field_size;
		if (i != dimension1 - 1)
		    printf(", ");
	    }
	    printf("}");
	}
	printf("}");
    }
    if (IOdumpVerbose)
	printf("; ");
    else
	printf(" ");
}

static void
byte_swap(data, size)
char *data;
int size;
{
    int i;
    assert((size % 2) == 0);
    for (i = 0; i < size / 2; i++) {
	char tmp = data[i];
	data[i] = data[size - i - 1];
	data[size - i - 1] = tmp;
    }
}

/*
** Code below this point relates to operations to and from the format server 
*/

static int
put_serverAtomicInt(fd, file_int_ptr, iofile)
void *fd;
FILE_INT *file_int_ptr;
IOFile iofile;
{
#if SIZEOF_INT == 4
    int tmp_value = *file_int_ptr;
    int junk_errno;
    char *junk_result_str;
    if (os_server_write_func(fd, &tmp_value, 4, &junk_errno, &junk_result_str) != 4)
	return 0;
#else
    Baaad shit;
#endif
    return 1;
}

static int
get_serverAtomicInt(fd, file_int_ptr, byte_reversal)
void *fd;
FILE_INT *file_int_ptr;
int byte_reversal;
{
#if SIZEOF_INT == 4
    int tmp_value;
    int junk_errno;
    char *junk_result_str;
    if (os_server_read_func(fd, &tmp_value, 4, &junk_errno, &junk_result_str) != 4)
	return 0;
#else
    Baaad shit;
#endif
    if (byte_reversal)
	byte_swap((char *) &tmp_value, 4);
    *file_int_ptr = tmp_value;
    return 1;
}

extern int
serverAtomicRead(fd, buffer, length)
void* fd;
void *buffer;
int length;
{
     char *junk_result_str;
     int junk_errno;
     return os_server_read_func(fd, buffer, length, &junk_errno,
				&junk_result_str);
}

static int
serverAtomicWrite(fd, buffer, length)
void* fd;
void *buffer;
int length;
{
     char *junk_result_str;
     int junk_errno;
     return os_server_write_func(fd, buffer, length, &junk_errno,
				&junk_result_str);
}


static void
provisional_use_warning(fd)
int fd;
{
    static int warned = 0;

    if (warned) return;
    warned++;

    fprintf(stderr, "The contacted format_server daemon allows only temporary use.\n");
    fprintf(stderr, " See http://www.cc.gatech.edu/systems/projects/MOSS/servers.html for more info.\n");
}	

static void
server_register_format(iofile, ioformat)
IOFile iofile;
IOFormat ioformat;
{
    int tries = 0;
    char format_registration_char = 'F';
    char return_char = 0;
    unsigned char *server_ID;

    if (iofile->master_context != NULL) {
	server_register_format((IOFile)iofile->master_context, ioformat);
	return;
    }

retry:
    if (tries++ > 2) {
	printf("Failed to contact format server\n");
	exit(1);
    }

    establish_server_connection_ptr(iofile, 1);

    if (serverAtomicWrite(iofile->server_fd, &format_registration_char, 1) != 1)
	goto retry;

    if (! write_format_to_fd(iofile->server_fd, ioformat)) 
	goto retry;

    serverAtomicRead(iofile->server_fd, &return_char, 1);
    if (return_char == 'P') {
	provisional_use_warning((int)(long)iofile->server_fd);
	serverAtomicRead(iofile->server_fd, &return_char, 1);
    }
    assert(return_char == 'I');
    if (serverAtomicRead(iofile->server_fd, (void*)&ioformat->server_ID, 
			 (int)sizeof(ioformat->server_ID)) 
	!= sizeof(ioformat->server_ID)) {
	goto retry;
    }
    server_ID = (unsigned char *)&ioformat->server_ID;
}

static char *
stringify_field_type(type, base_format, buffer, size)
char *type;
IOFormat base_format;
char *buffer;
int size;
{
    char *index_start;
    unsigned char *server_ID = (unsigned char*)&base_format->server_ID;

    if (base_format == NULL) return type;

    index_start = strchr(type, '[');
    assert(strlen(type) + 2*sizeof(base_format->server_ID +2) <= (unsigned int)size);
    if (index_start == NULL) {
	sprintf(buffer, "%s:%02x%02x%02x%02x:%02x%02x%02x%02x", type, 
		  server_ID[0], server_ID[1], server_ID[2], server_ID[3],
		  server_ID[4], server_ID[5], server_ID[6], server_ID[7]);
    } else {
	*index_start = 0;
	sprintf(buffer, "%s:%02x%02x%02x%02x:%02x%02x%02x%02x[%s", type, 
		server_ID[0], server_ID[1], server_ID[2], server_ID[3],
		server_ID[4], server_ID[5], server_ID[6], server_ID[7],
		(index_start +1));
	*index_start = '[';
    }
    return buffer;
}

extern
char *
global_name_of_IOformat(format)
IOFormat format;
{
    int size = strlen(format->format_name) + 3 + 
	2 *sizeof(format->server_ID);
    char *buffer = malloc(size);
    return stringify_field_type(format->format_name, format, buffer, size);
}

static void
unstringify_field_type(type, buffer, size)
char *type;
char *buffer;
int size;
{
    char *last_colon = strrchr(type, ':');
    char *second_last_colon;
    char *server_ID = (char*)buffer;
    int  int_server_ID[8];
    int  i;

    assert(size > 8);
    *last_colon = 0;
    second_last_colon = strrchr(type, ':');
    *last_colon = ':';
    assert(sscanf(second_last_colon, ":%02x%02x%02x%02x:%02x%02x%02x%02x", 
		  &int_server_ID[0], &int_server_ID[1],
		  &int_server_ID[2], &int_server_ID[3],
		  &int_server_ID[4], &int_server_ID[5],
		  &int_server_ID[6], &int_server_ID[7]) == 8);
    for (i=0; i<8; i++) {
      server_ID[i] = int_server_ID[i];
    }
}


#ifdef WORDS_BIGENDIAN
#define OUR_BYTE_ORDER 1
#define OTHER_BYTE_ORDER 0
#else
#define OUR_BYTE_ORDER 0
#define OTHER_BYTE_ORDER 1
#endif

extern int
write_format_to_fd(fd, ioformat)
void *fd;
IOFormat ioformat;
{
    FILE_INT name_length, field_count, record_length, format_id;
    int field;
    int result = 1;
    int record_byte_order;
    int pointer_size;
    int IOversion;

    if ((!ioformat) || (!ioformat->format_name)) {
	int zero = 0;
	result = put_serverAtomicInt(fd, &zero, NULL);
	return !result;
    }
    name_length = strlen(ioformat->format_name);
    format_id = ioformat->format_id;
    field_count = ioformat->field_count;
    record_length = ioformat->record_length;
    pointer_size = ioformat->pointer_size;
    IOversion = ioformat->IOversion;
    if (ioformat->byte_reversal) {
      record_byte_order = OTHER_BYTE_ORDER;
    } else {
      record_byte_order = OUR_BYTE_ORDER;
    }
    result &= put_serverAtomicInt(fd, &name_length, NULL);
    result &= put_serverAtomicInt(fd, &field_count, NULL);
    result &= put_serverAtomicInt(fd, &record_length, NULL);
    result &= put_serverAtomicInt(fd, &record_byte_order, NULL);
    result &= put_serverAtomicInt(fd, &pointer_size, NULL);
    result &= put_serverAtomicInt(fd, &IOversion, NULL);
    if (!result)
	return 0;
    if (serverAtomicWrite(fd, ioformat->format_name, name_length + 1) !=
	name_length + 1)
	return 0;

    for (field = 0; field < ioformat->field_count; field++) {
	FILE_INT field_name_len, field_type_len, field_size, field_offset;
	IOField *iofield = &(ioformat->field_list[field]);
	char field_type_buffer[512];
	char *ftype;

	memset(field_type_buffer, 0, sizeof(field_type_buffer));
	ftype = stringify_field_type(iofield->field_type,
				     ioformat->var_list[field].subformat,
				     field_type_buffer, 
				     (int)sizeof(field_type_buffer));
	field_name_len = strlen(iofield->field_name);
	field_type_len = strlen(ftype);
	field_size = iofield->field_size;
	field_offset = iofield->field_offset;
	result &= put_serverAtomicInt(fd, &field_name_len, NULL);
	result &= put_serverAtomicInt(fd, &field_type_len, NULL);
	result &= put_serverAtomicInt(fd, &field_size, NULL);
	result &= put_serverAtomicInt(fd, &field_offset, NULL);
	if (!result) return 0;
	if (serverAtomicWrite(fd, iofield->field_name, field_name_len +1)!=
	    field_name_len + 1)
	    return 0;
	if (serverAtomicWrite(fd, ftype, field_type_len +1)!=
	    field_type_len + 1)
	    return 0;
    }
    return 1;
}

static IOFormat
read_format_from_fd(fd, byte_reversal)
void *fd;
int byte_reversal;
{
    FILE_INT name_length;
    FILE_INT field_count;
    FILE_INT record_length;
    IOFormat ioformat = new_IOFormat();
    IOVarInfoList new_var_list;
    int field;
    int record_byte_order;
    int pointer_size;
    int IOversion;

    get_serverAtomicInt(fd, &name_length, byte_reversal);
    if (name_length == 0) {
	free(ioformat);
	return NULL;
    }
    get_serverAtomicInt(fd, &field_count, byte_reversal);
    get_serverAtomicInt(fd, &record_length, byte_reversal);
    get_serverAtomicInt(fd, &record_byte_order, byte_reversal);
    get_serverAtomicInt(fd, &pointer_size, byte_reversal);
    get_serverAtomicInt(fd, &IOversion, byte_reversal);

    ioformat->format_name = io_malloc(name_length + 1);
    ioformat->field_count = field_count;
    ioformat->variant = FALSE;
    ioformat->record_length = record_length;
    ioformat->native_format = 0;
    ioformat->conversion = NULL;
    ioformat->byte_reversal = (record_byte_order != OUR_BYTE_ORDER);
    ioformat->pointer_size = pointer_size;
    ioformat->IOversion = IOversion;
    serverAtomicRead(fd, ioformat->format_name, name_length + 1);
    ioformat->field_list = (IOFieldList) io_malloc(sizeof(IOField) *
					    (ioformat->field_count + 1));
    new_var_list = (IOVarInfoList)
	io_malloc((size_t) sizeof(IOVarInfoStruct) * field_count);
    ioformat->var_list = new_var_list;


    for (field = 0; field < ioformat->field_count; field++) {
	FILE_INT field_name_len;
	FILE_INT field_type_len;
	FILE_INT field_size;
	FILE_INT field_offset;
	IOField *iofield = &(ioformat->field_list[field]);
	ioformat->var_list[field].control_field = NULL;
	ioformat->var_list[field].subformat = NULL;
	get_serverAtomicInt(fd, &field_name_len, byte_reversal);
	get_serverAtomicInt(fd, &field_type_len, byte_reversal);
	get_serverAtomicInt(fd, &field_size, byte_reversal);
	get_serverAtomicInt(fd, &field_offset, byte_reversal);

	iofield->field_size = field_size;
	iofield->field_offset = field_offset;
	iofield->field_name = io_malloc(field_name_len + 1);
	iofield->field_type = io_malloc(field_type_len + 1);
	serverAtomicRead(fd, iofield->field_name, field_name_len + 1);
	serverAtomicRead(fd, iofield->field_type, field_type_len + 1);
	new_var_list[field].data_type = 
	    array_str_to_data_type(iofield->field_type, 
				   &new_var_list[field].dimen1, 
				   &new_var_list[field].dimen2);
	
    }
    return ioformat;
}

/*
 * server_read_format() is used by the format server to read a format 
 * being registered by a client.
 */
extern IOFormat
server_read_format(fs, fsc)
format_server fs;
FSClient fsc;
{
    return read_format_from_fd(fsc->fd, fsc->byte_reversal);
}

static IOFormat
server_get_format(iocontext, buffer)
IOContext iocontext;
void *buffer;
{
    IOFile iofile = (IOFile) iocontext;
    IOFormat ioformat;
    IOFieldList field_list;
    char format_get_char = 'G';
    char return_char = 0;
    int id_size = sizeof(iofile->format_list[0]->server_ID);
    int field;

    establish_server_connection_ptr(iofile, 1);

    if (serverAtomicWrite(iofile->server_fd, &format_get_char, 1) != 1) {
	printf("write failed\n");
	return NULL;
    }
    if (serverAtomicWrite(iofile->server_fd, buffer, id_size) != id_size) {
	printf("write failed\n");
	return NULL;
    }
    serverAtomicRead(iofile->server_fd, &return_char, 1);
    if (return_char == 'P') {
	provisional_use_warning((int)(long)iofile->server_fd);
	serverAtomicRead(iofile->server_fd, &return_char, 1);
    }
    assert(return_char == 'F');
    ioformat = read_format_from_fd(iofile->server_fd, 
				   iofile->server_byte_reversal);
    
    if (iofile->reg_format_count == iofile->format_list_size) {
	expand_IOFile(iofile);
    }
    if (ioformat == NULL) return NULL;

    ioformat->format_id = iofile->reg_format_count++;
    iofile->format_list[ioformat->format_id] = ioformat;
    ioformat->iofile = iofile;
    ioformat->record_length = 0;
    ioformat->variant = FALSE;
    memcpy(ioformat->server_ID, buffer, id_size);
    field_list = ioformat->field_list;
    for (field = 0; field < ioformat->field_count; field++) {
	int field_size = 0;
	if (get_var_array_control(field_list[field].field_type, field_list)
	    != -1) {
	    /* variant array, real_field_size is ioformat->pointer_size */
	    field_size = ioformat->pointer_size;
	} else {
	    int dimen1 = 1, dimen2 = 1;
	    IOdata_type base_type;
	    base_type = array_str_to_data_type(field_list[field].field_type,
					       &dimen1, &dimen2);
	    if ((base_type != unknown_type) &&
		(field_list[field].field_size > 16)) {
		fprintf(stderr, "Field size for field %s in format %s is large, check to see if it is valid.\n",
		    field_list[field].field_name, ioformat->format_name);
	    }
	    field_size = field_list[field].field_size * dimen1 * dimen2;
	}
	assert(field_size > 0);
	ioformat->record_length = Max(ioformat->record_length,
				      field_list[field].field_offset +
				      field_size);
    }
    generate_var_list(ioformat);
    for (field = 0; field < ioformat->field_count; field++) {
	if (ioformat->var_list[field].string == TRUE) {
	    ioformat->variant = TRUE;
	} else {
	    char *base_type =
	      base_data_type(field_list[field].field_type);
	    IOFormat subformat = NULL;

	    /* if field is of another record format, fill that in */
	    if (str_to_data_type(base_type) == unknown_type) {
		subformat = get_IOformat_by_name_IOcontext((IOContext)iofile, 
							   base_type);
		ioformat->var_list[field].subformat = subformat;
	    }
	    if (ioformat->var_list[field].var_array == TRUE) {
		/* got a variant array */
		ioformat->variant = 1;
	    } else {
		/* if field is variant by its subformat being variant */
		if (subformat != NULL) {
		    ioformat->variant |= subformat->variant;
		}
	    }
	    free(base_type);
	}
    }
    return ioformat;
}

extern void
server_get_server_ID(fd, server_ID)
void *fd;
void *server_ID;
{
    int id_size = sizeof(((IOFile)0)->format_list[0]->server_ID);

    serverAtomicRead(fd, server_ID, id_size);
}

extern void
print_format_ID(ioformat)
IOFormat ioformat;
{
    int id_size = sizeof(((IOFile)0)->format_list[0]->server_ID);
    int i;
    unsigned char *ID = (unsigned char *)&ioformat->server_ID[0];
    for (i=0; i < id_size; i++) {
	printf("%2x", ID[i]);
    }
}

/* write header information to the format server */
extern void
server_write_header(iofile)
IOFile iofile;
{
    FILE_INT magic = MAGIC_NUMBER+1;
    FILE_INT float_format = Format_IEEE;
    put_serverAtomicInt(iofile->server_fd, &magic, iofile);
    put_serverAtomicInt(iofile->server_fd, &float_format, iofile);

    get_serverAtomicInt(iofile->server_fd, &magic, 0);

    if (magic != MAGIC_NUMBER) {
	if (magic == REVERSE_MAGIC_NUMBER) {
	    iofile->server_byte_reversal = 1;
	} else {
	    /* close client */
	    return;
	}
    }
}

extern void
server_write_char(fsc, chr)
FSClient fsc;
char *chr;
{
    serverAtomicWrite(fsc->fd, chr, 1);
}

extern void
server_read_header(fsc)
FSClient fsc;
{
    FILE_INT magic;
    FILE_INT float_format;

    fsc->byte_reversal = 0;
    get_serverAtomicInt(fsc->fd, &magic, fsc->byte_reversal);

    if (magic != MAGIC_NUMBER) {
	if (magic == REVERSE_MAGIC_NUMBER) {
	    fsc->byte_reversal = 1;
	} else {
	    if (magic == (MAGIC_NUMBER + 1)) {
		fsc->new_version++;
	    } else if (magic == (REVERSE_MAGIC_NUMBER + 0x1000000)) {
		fsc->byte_reversal = 1;
		fsc->new_version++;
	    } else {
		/* close client */
		return;
	    }
	}
    }
    get_serverAtomicInt(fsc->fd, &float_format, fsc->byte_reversal);
    magic = MAGIC_NUMBER;
    put_serverAtomicInt(fsc->fd, &magic, NULL);
    if (float_format != Format_IEEE) {
	assert(FALSE);
    }
}
